# mypy: allow_untyped_calls

import sys

if sys.version_info < (3, 8):
    from typing_extensions import Final
    from importlib_metadata import version
else:
    from typing import Final
    from importlib.metadata import version

__version__: Final[str] = version(__name__)

from flake8_force_keyword_arguments.checker import Checker  # noqa: E402



import ast
import inspect
from enum import Enum
from types import ModuleType
from typing import Any, Callable, Iterable, Optional, Set


class QualifierOption(str, Enum):
    ONLY_NAME = 'only_name'
    ONLY_WITH_QUALIFIER = 'only_with_qualifier'
    BOTH = 'both'


def list_pos_only_callables(
    m: ModuleType,
    parent_module_qualifier: str,
    poa_threshold: int,
    qualifier_option: QualifierOption = QualifierOption.BOTH,
) -> Iterable[str]:
    """POA: Positional Only Arguments"""
    return _list_pos_only_callables(
        m=m,
        parent_module_qualifier=parent_module_qualifier,
        poa_threshold=poa_threshold,
        visited=set(),
        qualifier_option=qualifier_option,
    )


def _list_pos_only_callables(
    m: ModuleType,
    parent_module_qualifier: str,
    poa_threshold: int,
    visited: Set[ModuleType],
    qualifier_option: QualifierOption = QualifierOption.BOTH,
) -> Iterable[str]:
    """POA: Positional Only Arguments"""
    if m in visited:
        return

    visited.add(m)

    parent_module_qualifiers = parent_module_qualifier.split('.')
    for name, obj in inspect.getmembers(m):
        if name.startswith('_'):
            continue

        if inspect.ismodule(obj):
            yield from _list_pos_only_callables(
                m=obj,
                parent_module_qualifier=f'{parent_module_qualifier}.{name}',
                poa_threshold=poa_threshold,
                visited=visited,
                qualifier_option=qualifier_option,
            )

        if not does_callable_have_poa_more_than(o=obj, poa_threshold=poa_threshold):
            continue

        if parent_module_qualifier == 'builtins' or qualifier_option is not QualifierOption.ONLY_WITH_QUALIFIER:
            yield name

        if qualifier_option is not QualifierOption.ONLY_NAME:
            for i in range(len(parent_module_qualifiers)):
                yield f'{".".join(parent_module_qualifiers[i:])}.{name}'


def does_callable_have_poa_more_than(o: object, poa_threshold: int) -> bool:
    """POA: Positional Only Arguments"""
    func = get_inspectable_function(o)
    if func is None:
        return False

    sig = inspect.signature(func)
    params = tuple(sig.parameters.values())
    if inspect.ismethoddescriptor(func) and len(params) > 0:
        params = params[1:]

    poa_count = 0
    for p in params:
        if p.kind is inspect.Parameter.VAR_POSITIONAL:
            return True
        elif p.kind is inspect.Parameter.POSITIONAL_ONLY:
            poa_count += 1

    return poa_count >= poa_threshold


def get_inspectable_function(o: object) -> Optional[Callable[..., Any]]:
    try:
        inspect.signature(o)  # type: ignore[arg-type]
    except (ValueError, TypeError):
        pass
    else:
        return o  # type: ignore

    if inspect.isclass(o):
        try:
            inspect.signature(o.__init__)  # type: ignore[misc]
        except (ValueError, TypeError):
            return None
        else:
            return o.__init__  # type: ignore[no-any-return,misc]
    elif callable(o):
        try:
            inspect.signature(o.__call__)
        except (ValueError, TypeError):
            return None
        else:
            return o.__call__
    return None


def get_invocation_line(c: ast.Call) -> str:
    def dfs(a: ast.AST) -> str:
        child = getattr(a, 'value', None)
        name = getattr(a, 'id', getattr(a, 'attr', None))

        if child is None or not isinstance(child, ast.AST):
            if name is None or not isinstance(name, str):
                return ''
            return name

        return '.'.join(filter(None, (dfs(child), name)))

    return dfs(c.func)



import ast
import importlib
import re
from argparse import Namespace
from typing import ClassVar, Iterable, Tuple, Type

import sys
from flake8.options.manager import OptionManager
from itertools import chain
from marisa_trie import Trie

import flake8_force_keyword_arguments
from flake8_force_keyword_arguments import util

if sys.version_info < (3, 8):
    from typing_extensions import Final
else:
    from typing import Final

if sys.version_info < (3, 9):
    from typing import Pattern
else:
    from re import Pattern

DEFAULT_MAX_POS_ARGS: Final[int] = 2
DEFAULT_IGNORE_PATTERNS: Final[
    str
] = r'(:?^logger.(:?log|debug|info|warning|error|exception|critical)$|__setattr__$|__delattr__$|__getattr__$)'
DEFAULT_INSPECT_MODULES: Final[Tuple[str, ...]] = ('builtins',)
DEFAULT_QUALIFIER_OPTION: Final[util.QualifierOption] = util.QualifierOption.BOTH


class Checker:
    name: Final[str] = flake8_force_keyword_arguments.__name__
    version: Final[str] = flake8_force_keyword_arguments.__version__
    MESSAGE_TEMPLATE: Final[
        str
    ] = 'FKA100 {function_name}\'s call uses {number_of_args} positional arguments, use keyword arguments.'

    _max_pos_args: ClassVar[int] = DEFAULT_MAX_POS_ARGS
    _ignore_patterns: ClassVar[Tuple[Pattern[str], ...]] = ()
    _ignore_trie: ClassVar[Trie]

    _tree: ast.AST

    def __init__(self, tree: ast.AST) -> None:
        self._tree = tree

    @classmethod
    def add_options(cls, parser: OptionManager) -> None:
        parser.add_option(  # pragma: no cover
            '--kwargs-max-positional-arguments',
            type=int,
            dest='max_positional_arguments',
            default=DEFAULT_MAX_POS_ARGS,
            parse_from_config=True,
            help='How many positional arguments are allowed (default: %(default)s)',
        )
        parser.add_option(  # pragma: no cover
            '--kwargs-ignore-function-pattern',
            type=str,
            dest='ignore_function_pattern',
            default=DEFAULT_IGNORE_PATTERNS,
            parse_from_config=True,
            help='Ignore pattern list (default: %(default)s)',
        )
        parser.add_option(  # pragma: no cover
            '--kwargs-ignore-function-pattern-extend',
            type=str,
            dest='ignore_function_pattern_extend',
            default=None,
            parse_from_config=True,
            help='Extend ignore pattern list.',
        )
        parser.add_option(  # pragma: no cover
            '--kwargs-inspect-module',
            dest='inspect_module',
            comma_separated_list=True,
            default=DEFAULT_INSPECT_MODULES,
            parse_from_config=True,
            help=(
                'Inspect module level constructor of classes or functions to gather positional only callables '
                'and ignore it on lint. Note that methods are not subject to inspection. (default: %(default)s)'
            ),
        )
        parser.add_option(  # pragma: no cover
            '--kwargs-inspect-module-extend',
            dest='inspect_module_extend',
            comma_separated_list=True,
            default=(),
            parse_from_config=True,
            help='Extend --kwargs-inspect-module',
        )
        parser.add_option(  # pragma: no cover
            '--kwargs-inspect-qualifier-option',
            type=util.QualifierOption,
            dest='inspect_qualifier_option',
            choices=tuple(map(lambda v: v.value, util.QualifierOption.__members__.values())),
            default=DEFAULT_QUALIFIER_OPTION.value,
            parse_from_config=True,
            help=(
                'For detected positional only callables by inspection, option to append the qualifier or not. '
                'e.g. In case builtins.setattr(), '
                '`both` will register `builtins.setattr` and `setattr` as positional only function. '
                '`only_name` will register `setattr` and `only_with_qualifier` will register `builtins.setattr`. '
                '(default: %(default)s)'
            ),
        )

    @classmethod
    def parse_options(cls, options: Namespace) -> None:
        cls._max_pos_args = options.max_positional_arguments

        ignore_patterns = (options.ignore_function_pattern, options.ignore_function_pattern_extend)
        cls._ignore_patterns = tuple(map(re.compile, filter(None, ignore_patterns)))

        qualifier_option = options.inspect_qualifier_option

        cls._ignore_trie = Trie(
            chain.from_iterable(
                map(
                    lambda module_name: util.list_pos_only_callables(
                        m=importlib.import_module(module_name),
                        parent_module_qualifier=module_name,
                        poa_threshold=cls._max_pos_args,
                        qualifier_option=qualifier_option,
                    ),
                    chain(options.inspect_module, options.inspect_module_extend),
                )
            )
        )

    def run(self) -> Iterable[Tuple[int, int, str, Type['Checker']]]:
        for node in ast.walk(self._tree):
            if not isinstance(node, ast.Call) or len(node.args) < self._max_pos_args:
                continue

            invocation_line = util.get_invocation_line(node)

            # ignored because of patterns
            if any(map(lambda p: p.search(invocation_line), self._ignore_patterns)):
                continue

            # ignored because of inspection
            if invocation_line in self._ignore_trie:
                continue

            message = self.MESSAGE_TEMPLATE.format(function_name=invocation_line, number_of_args=len(node.args))
            yield node.lineno, node.col_offset, message, type(self)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flake8_force_keyword_arguments']

package_data = \
{'': ['*']}

install_requires = \
['flake8>=3.8', 'marisa-trie>=0.7.7,<0.8.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata', 'typing-extensions']}

entry_points = \
{'flake8.extension': ['FKA1 = flake8_force_keyword_arguments:Checker']}

setup_kwargs = {
    'name': 'flake8-force-keyword-arguments',
    'version': '1.0.2',
    'description': 'A flake8 extension that is looking for function calls and forces to use keyword arguments if there are more than X arguments',
    'long_description': "# flake8-force-keyword-arguments\n\n[![PyPI](https://img.shields.io/pypi/v/flake8-force-keyword-arguments?label=pypi&logo=pypi&style=flat-square)](https://pypi.org/project/flake8-force-keyword-arguments/)\n[![PyPI - Wheel](https://img.shields.io/pypi/wheel/flake8-force-keyword-arguments?style=flat-square&logo=pypi)](https://pypi.org/project/flake8-force-keyword-arguments/)\n[![Python Version](https://img.shields.io/pypi/pyversions/flake8-force-keyword-arguments.svg?style=flat-square&logo=python)](https://pypi.org/project/flake8-force-keyword-arguments/)\n[![PyPI - Implementation](https://img.shields.io/pypi/implementation/flake8-force-keyword-arguments?style=flat-square&logo=python)]((https://pypi.org/project/flake8-force-keyword-arguments/))\n![Codecov](https://img.shields.io/codecov/c/gh/isac322/flake8-force-keyword-arguments?style=flat-square&logo=codecov)\n![GitHub last commit (branch)](https://img.shields.io/github/last-commit/isac322/flake8-force-keyword-arguments/master?logo=github&style=flat-square)\n![GitHub Workflow Status (branch)](https://img.shields.io/github/workflow/status/isac322/flake8-force-keyword-arguments/CI/master?logo=github&style=flat-square)\n![Dependabot Status](https://flat.badgen.net/github/dependabot/isac322/flake8-force-keyword-arguments?icon=github)\n[![Code Style](https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square)](https://github.com/psf/black)\n___\n\nA flake8 plugin that is looking for function calls and forces to use keyword arguments\nif there are more than X (default=2) arguments.\nAnd it can ignore positional only or variable arguments functions such as `setattr()` or `Logger.info()`.\nThe plugin inspects given modules (via `--kwargs-inspect-module`, `--kwargs-inspect-module-extend`)\nto get signature and to determine whether it is positional only or variable arguments function.\nThe inspection runs only once at the start of the flake8 command and remembers ignore list through runtime.\n\n\n## Installation\n\n```\npip install flake8-force-keyword-arguments\n```\n\n## Usage\n\nRun your `flake8` checker [as usual](http://flake8.pycqa.org/en/latest/user/invocation.html).\n\nExample:\n\n```bash\nflake8 your_module.py\n```\n\n## Option\n\n- `--kwargs-max-positional-arguments`: How many positional arguments are allowed (default: 2)\n- `--kwargs-ignore-function-pattern`: Ignore pattern list (default: ('^logger.(:?log|debug|info|warning|error|exception|critical)$', '__setattr__$', '__delattr__$', '__getattr__$'))\n- `--kwargs-ignore-function-pattern-extend`: Extend ignore pattern list.\n- `--kwargs-inspect-module`: Inspect module level constructor of classes or functions to gather positional only callables and ignore it on lint. Note that methods are not subject to inspection. (default: ('builtins',))\n- `--kwargs-inspect-module-extend`: Extend `--kwargs-inspect-module`\n- `--kwargs-inspect-qualifier-option {only_name,only_with_qualifier,both}`: For detected positional only callables by inspection, option to append the qualifier or not. e.g. In case builtins.setattr(), `both` will register `builtins.setattr` and `setattr` as positional only function. `only_name` will register `setattr` and `only_with_qualifier` will register `builtins.setattr`. (default: QualifierOption.BOTH)\n\n## Example\n\n### code: `test.py`\n\n```python\nfrom functools import partial\n\ndef one_argument(one):\n    pass\n\ndef two_arguments(one, two):\n    pass\n\ndef pos_only_arguments(one, two, three, /):  # python 3.8 or higher required\n    pass\n\ndef variable_arguments(*args):\n    pass\n\nvariadic = lambda *args: None\ncurried = partial(variadic, 1)\n\none_argument(1)\none_argument(one=1)\ntwo_arguments(1, 2)\ntwo_arguments(one=1, two=2)\npos_only_arguments(1, 2, 3)\nvariadic(1, 2, 3)\ncurried(2, 3)\n```\n\n### Command\n\n#### `flake8 test.py --select FKA1 --kwargs-inspect-module-extend test`\n\n```\ntest.py:20:1: FKA100 two_arguments's call uses 2 positional arguments, use keyword arguments.\n```\n\n#### `flake8 test.py --select FKA1 --kwargs-inspect-module-extend test --kwargs-ignore-function-pattern-extend ^two_arguments$`\n\nNo error\n\n#### `flake8 test.py --select FKA1`\n\n```\ntest.py:16:11: FKA100 partial's call uses 2 positional arguments, use keyword arguments.\ntest.py:20:1: FKA100 two_arguments's call uses 2 positional arguments, use keyword arguments.\ntest.py:22:1: FKA100 pos_only_arguments's call uses 3 positional arguments, use keyword arguments.\ntest.py:23:1: FKA100 variadic's call uses 3 positional arguments, use keyword arguments.\ntest.py:24:1: FKA100 curried's call uses 2 positional arguments, use keyword arguments.\n```\n\n## Limitation\n\nCurrently it only inspects given modules and can not inspect (static, class or normal) methods.\nBecause inspection carries import, it is not safe to inspect all possible packages.\nAnd method case, the plugin can inspect methods signature and also can determine whether it is positional only or not.\nBut it can not use the information on lint time.\nBecause python is a dynamic typed language and flake8 is basically a static analyzer.\nThat is, flake8 can not get type information of `logger.debug()`.\nSo even if I know that `logging.Logger::debug()` is a variadic function,\nI can not assure that `logger` is a instance of `Logger`.\n\n## Error codes\n\n| Error code |                     Description                                |\n|:----------:|:--------------------------------------------------------------:|\n|   FKA100    | XXX's call uses N positional arguments, use keyword arguments. |\n",
    'author': 'Viktor Chaptsev',
    'author_email': 'viktor@chaptsev.ru',
    'maintainer': 'Byeonghoon Yoo',
    'maintainer_email': 'bh322yoo@gmail.com',
    'url': 'https://github.com/isac322/flake8-force-keyword-arguments',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
